#==============================================================================

# Copyright (c) 2016-10-21, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

# Inspired by
# http://scikit-learn.org/stable/auto_examples/decomposition/plot_faces_decomposition.html

#==============================================================================

import numpy as np
import matplotlib.pyplot as plt
import time

from sklearn.datasets import fetch_olivetti_faces
from sklearn import decomposition

IMAGE_SHAPE = (64, 64)
SHOW_BLOCKS = True # Program should stop at each figure or run through.
CENTER_FACES = True # Faces should be centered or not.

#==============================================================================

# PLOTTING ROUTINES

def show_gallery(title, images, num_col=1, num_row=1):
    """
    Routine to plot a set of square images.
    """
 
    unit = 1.8 # Change this unit if you want the figures larger or smaller.

    top_space    = 0.40*unit
    left_space   = 0.10*unit
    bottom_space = 0.10*unit
    right_space  = left_space
    image_height = 1.00*unit
    image_hspace = 0.05*unit
    image_width  = image_height
    image_wspace = image_hspace

    fig_height = num_row*image_height + (num_row-1)*image_hspace +  top_space + bottom_space
    fig_width  = num_col*image_width  + (num_col-1)*image_wspace + left_space +  right_space

    fig = plt.figure(figsize=(fig_width, fig_height))
    fig.suptitle(title, size=int(15*unit))
    for i, image in enumerate(images):
        ax = fig.add_subplot(num_row, num_col, i + 1)
        ax.imshow(image.reshape(IMAGE_SHAPE), cmap=plt.cm.gray,
                   interpolation='nearest'
        )
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])
    fig.subplots_adjust(left=left_space/fig_width, bottom=bottom_space/fig_height,
                        right=1-right_space/fig_width, top=1-top_space/fig_height,
                        wspace=(num_row-1)*image_wspace/fig_width,
                        hspace=(num_col-1)*image_hspace/fig_height)
    plt.show(block=SHOW_BLOCKS)


def animate_gallery(images, stride=1):
    """
    Routine to animate a set of images.
    """

    fig = plt.figure(figsize=(2.,2.26))
    fig.suptitle("Animation of faces", size=16)
    plt.show(block=False)
    ax = fig.add_subplot(111)
    for i, image in enumerate(images):
        if not i%stride == 0:
            continue
        ax.imshow(image.reshape(IMAGE_SHAPE), cmap=plt.cm.gray,
                  interpolation='nearest')
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])
        fig.canvas.draw()    
    plt.show(block=SHOW_BLOCKS)


def animate_proj(faces_proj, comp_1=0, comp_2=1, sleep=0.2):
    """
    Routine to animate projected points of one id within all points of all ids.
    """

    fig = plt.figure(figsize=(10,10))
    fig.suptitle("Animation of projected face images in PCA components "+str(comp_1)+" and "+str(comp_2), size=16)
    plt.show(block=False)
    ax = fig.add_subplot(111)
    for id_idx in range(40):
        ax.clear()
        ax.scatter(faces_proj[:,comp_1],
                   faces_proj[:,comp_2],
                   s=10, marker='.')
        ax.scatter(faces_proj[id_idx*10:(id_idx+1)*10,comp_1],
                   faces_proj[id_idx*10:(id_idx+1)*10,comp_2], 
                   s=50, color='r', marker='o')
        fig.canvas.draw()
        time.sleep(sleep)
    plt.show(block=SHOW_BLOCKS)


#==============================================================================

# LOAD FACE IMAGES

# The Olivetti face dataset used here consists of 10 slightly different views
# or expressions of 40 faces, each image with 64 x 64 pixels.  The images are
# aligned at the pupils of the eyes.  They are stored in a 400 x 4096 array.

faces = fetch_olivetti_faces().data
print "faces.shape =", faces.shape

# Plot all 10 views of the first five faces in the data base.
print "Show all raw images of first five individuals in an overview plot."
show_gallery("All views of first five faces", faces[:50], num_col=10, num_row=5)

# Animate gallery of face to see how well aligned the faces are. 
print "Show all raw images of first five individuals in an animation plot to see the alignment at the eyes."
animate_gallery(faces, 5)

#==============================================================================

# EXERCISES

# Explore the data with the help of PCA.  Start without looking at the
# exercises below.  Try to be creative here.  If you run out of ideas, look
# below.

#------------------------------------------------------------------------------

# EXERCISE 1: Center the face images by subtracting the mean face image.

# SOLUTION:
if CENTER_FACES:
    # faces_mean = ...
    print "faces_mean.shape =", faces_mean.shape # = (4096,)
    faces_repr = faces - faces_mean
    print "faces_repr.shape =", faces_repr.shape # = (400, 4096)

#------------------------------------------------------------------------------

# EXERCISE 2: When doing PCA, one normally centers the data first.  However, in
# some cases one might not want to do that and want the eigenvectors of the
# second moment matrix rather than of the covariance matrix.  The algorithm
# used here, however, does not seem to allow that, because it seems to
# implicitly or explicitly center the data itself (not documented).  What can
# you do to get the eigenvectors of the second moment matrix nevertheless?
# Implement your idea.

# SOLUTION:
if not CENTER_FACES:
    # This dummy faces_mean is needed for the code further down.
    faces_mean = np.zeros(4096)
    print "faces_mean.shape =", faces_mean.shape # = (4096,)
    # faces_repr = ...

#------------------------------------------------------------------------------

# EXERCISE 3: 
# (a): Visualize how faces look on the way between the mean face and the
# actual face.  
# (b): What do you think, which version looks prettiest?
# (c): Can you create caricatures by exaggerating the distance of a face from
# the mean face?

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 4: 
# (a): How many non-zero eigenvalues do you expect if you perform PCA on the
# centered face images?  Why?
# (b): What does that mean for the eigenvectors?

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 5: 
# (a): Perform PCA on the centered face images.  
# (b): How many non-zero eigenvalues do you get?

# The eigenvectors of face images are often called eigenfaces.

# SOLUTION:
# Hint: See http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
# eigenfaces = ... 
# eigenvalues = ... 
# ...
print "eigenfaces.shape =", eigenfaces.shape # = (400, 4096)
print "eigenvalues.shape =", eigenvalues.shape # = (400,)

#------------------------------------------------------------------------------

# EXERCISE 6: The PCA algorithm used here does not seem to determine more
# eigenvectors than there are data points.  Develop ideas (no need to implement
# that) how one could get the missing eigenvectors to get a full orthonormal
# basis?

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 7: 
# (a) Plot the eigenvalue spectrum.
# (b) How many eigenfaces, do you think, are useful?
# (c) Would the eigenvalue spectrum look much different if you had many more
# face images?

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 8: Plot the first 50 eigenfaces.

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 9: Show the effect of an eigenface on the mean face, i.e. visualize
# the mean face plus different amounts of an eigenface.  Do this for different
# eigenfaces.

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 10: Compare the eigenfaces with and without centering the face
# images.  What do you observe?  Explain the differences and similarities.

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 11: Project face images into the space of eigenvectors.

# SOLUTION:
# faces_proj = ...
print "faces_proj.shape =", faces_proj.shape # = (400, 4096)

#------------------------------------------------------------------------------

# EXERCISE 12: Reconstruct the faces from the projections.

# SOLUTION:
# faces_rec = ...
print "faces_rec.shape =", faces_rec.shape # = (400, 4096)

#------------------------------------------------------------------------------

# EXERCISE 13: Plot all 10 views of the first five reconstructed faces in the
# data base.

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 14: Show how the reconstruction depends on the number of eigenfaces used.

# SOLUTION:
# ...

#------------------------------------------------------------------------------

# EXERCISE 15: Animate the face images of individual ids (=persons) projected
# onto the first two eigenfaces to see how they cluster per id.

# SOLUTION:
# animate_proj(...)

#------------------------------------------------------------------------------

# EXERCISE 16: Imagine you have only one view per person in the gallery (but
# you were allowed to train PCA on all images for lack of more images).  Use
# the other views as probe images to perform face recognition.  A simple face
# recognition system based on nearest neighbor classification would search for
# the gallery image with the smallest Eucleadian distance to the probe image.
# Implement such a face recognition system on the raw images and the projected
# images.  Evaluate the performance and see how it depends on the number of
# eigenfaces used for the projection.  Maybe also check, how performance
# increases if you use more than one view per person in the gallery.

# SOLUTION:
# ...

#==============================================================================
#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

